#include <cmath>
#include "shared.h"
#include "analysisFunction.h"
#include "general.h"
class majorminor:public general{
public:
  //none optional stuff
  //  FILE *outfile;
  int doMajorMinor;
  int doCounts;
  int doMaf;
  int doAsso;
  int doCovar;
  int GL;
  void printArg(FILE *argFile);
  void run(funkyPars *pars);
  void clean(funkyPars *pars);  
  void print(funkyPars *pars);  
  void addDefault(funkyPars *pars);
  void openfile(const char *outfiles);
  majorminor(const char *outfiles,argStruct *arguments,int inputtype);
  void getOptions(argStruct *arguments);
  ~majorminor();
  //other stuff
  
};
void majorminor::printArg(FILE *argFile){
  fprintf(argFile,"%s:\n\n",__FILE__);
  fprintf(argFile,"-doMajorMinor\t%d\n",doMajorMinor);
  fprintf(argFile,"\t->1: Infer major and minor from GL\n");
  fprintf(argFile,"\t->2: Infer major and minor from allele counts\n");
  fprintf(argFile,"\n");

}

void majorminor::getOptions(argStruct *arguments){
  //default
  doMajorMinor=0;
  doMaf=0;
  doAsso=0;
  GL=0;
  //from command line
  doCounts=0;
  doMajorMinor=angsd::getArg("-doMajorMinor",doMajorMinor,arguments);
  if(doMajorMinor==-999){
    doMajorMinor=0;
    printArg(stdout);
    exit(0);
 
  }
  if(doMajorMinor==3){
    //    fprintf(stderr,"you should set -filter for this option\n");

  }
  
   doCounts=angsd::getArg("-doCounts",doCounts,arguments);
   
 

  if(doMajorMinor==0)
    return;

  doMaf=angsd::getArg("-doMaf",doMaf,arguments);
  doAsso=angsd::getArg("-doAsso",doAsso,arguments);
  doCovar=angsd::getArg("-doCovar",doCovar,arguments);
  GL=angsd::getArg("-GL",GL,arguments);
  printArg(arguments->argumentFile);
}

majorminor::majorminor(const char *outfiles,argStruct *arguments,int inputtype){
  //should be changes so that it is a command options
 getOptions(arguments);
 if(doMajorMinor==0)
    return;

 if(inputtype!=0&&inputtype!=5&&(doMaf||doAsso)){
   //   fprintf(stdout,"Need to infer major minor, setting doMajorMinor to 1\n");
   //  doMajorMinor=1;

 }
  /*
    inputtype 
    0) soap
    1) samglf 
    2) samglfClean 
    3) tglf 
    4) sim type 
    5) beagle
    6) mpileup
    7) bam
  */
 if(inputtype==5&&doMajorMinor){
   fprintf(stdout,"Error: Cannot estimate the major and minor based on posterior probabilities\n");
   exit(0);
 }
  if((inputtype==0 || inputtype==6 || inputtype==7)&&GL==0&&doMajorMinor==1){
    fprintf(stderr,"Error: For sequence data (bam,SOAP) likehoods (-GL) must be specified for major minor estimates (-doMajorMinor 1)\n");
    exit(0);
  }
  if(doMajorMinor==2&&doCounts==0){
    fprintf(stderr,"Error: -doCounts 1 needed for -doMajorMinor 2\n");
    exit(0);

  }


 //  fprintf(stderr,"Estimating major minor: %s\n",outfiles);
}

majorminor::~majorminor(){
  if(doMajorMinor==0)
    return;
}

void majorminor::clean(funkyPars *pars){
  //cleaning moved to printfunky
  if(doMajorMinor==0)
    return;
}

void majorminor::print(funkyPars *pars){
  //cleaning moved to printfunky
  if(doMajorMinor==0)
    return;
}

void majorminor::run(funkyPars *pars){
  if(doMajorMinor!=1)//this is for using likes for estimating major/minor
    return;
  if(pars->likes==NULL){
    fprintf(stderr,"majorminor really wants likelhioods\n");
    exit(0);
  }
  if(pars->minor!=NULL&&pars->major!=NULL){
    // return;    

  }else{

    pars->major = new int [pars->numSites];
    pars->minor = new int [pars->numSites];
    for(int i=0;i<pars->numSites;i++){
      pars->major[i] = 4;
      pars->minor[i] = 4;
    }
  }
  float lmax;
  float totalLike;
  int choiceMajor;
  int choiceMinor;
  for(int s=0;s<pars->numSites;s++) {
    int keepInd =0;// effective sample size (infered from GLS)
    for(int i=0;i<pars->nInd;i++){
      for(int ii=0;ii<10;ii++)
	if(pars->likes[s][i*10+ii]!=0){
	  keepInd++;
	  break;
	}
    }
    pars->results->freq->keepInd[s] = keepInd;
    if(keepInd==0){
      pars->keepSites[s] = 0;
      pars->major[s] = 4;
      pars->major[s] = 4;
      continue;
    }

    //if we had a read then estimate the major/minor
    lmax=-100000000;
    for(int Imajor=0;Imajor<3;Imajor++) {
      for(int Iminor=(Imajor+1);Iminor<4;Iminor++){
	//fprintf(stderr,"%d %d\n",Imajor,Iminor);
	totalLike=0;
	for(int i=0;i<pars->nInd;i++)
	  totalLike+=angsd::addProtect3(pars->likes[s][i*10+angsd::majorminor[Imajor][Imajor]]+log(0.25),
				 pars->likes[s][i*10+angsd::majorminor[Imajor][Iminor]]+log(0.5),
				 pars->likes[s][i*10+angsd::majorminor[Iminor][Iminor]]+log(0.25)
				 );
	if(totalLike>lmax){
	  lmax=totalLike;
	  choiceMajor=Imajor;
	  choiceMinor=Iminor;
	}
      }
    }
    float W0;
    float W1;
    float W2;
    float sum=0;
    for(int i=0;i<pars->nInd;i++){
      W0=exp(pars->likes[s][i*10+angsd::majorminor[choiceMajor][choiceMajor]])*0.25;
      W1=exp(pars->likes[s][i*10+angsd::majorminor[choiceMajor][choiceMinor]])*0.5;
      W2=exp(pars->likes[s][i*10+angsd::majorminor[choiceMinor][choiceMinor]])*0.25;
      sum+=(W1+2*W2)/(2*(W0+W1+W2));
      //      fprintf(stderr,"W %f\t%f\t%f\n",W0,W1,W2);
    }
    if(sum/pars->nInd<0.5){
      pars->major[s]=choiceMajor;
      pars->minor[s]=choiceMinor;
    }
    else{
      pars->major[s]=choiceMinor;
      pars->minor[s]=choiceMajor;
    }
    //    fprintf(stderr,"%d %d\n",pars->major[s],pars->minor[s]);
  } 
}
